@@ -54,6 +54,7 @@ const Meta = imports.gi.Meta;
 const St = imports.gi.St;
 
 const Applet = imports.ui.applet;
+const AppletManager = imports.ui.appletManager;
 const DND = imports.ui.dnd;
 const Main = imports.ui.main;
 const Panel = imports.ui.panel;
@@ -64,9 +65,10 @@ const Tooltips = imports.ui.tooltips;
 const Tweener = imports.ui.tweener;
 const Util = imports.misc.util;
 
-const DEFAULT_ICON_SIZE = 16; // too bad this can't be defined in theme (cinnamon-app.create_icon_texture returns a clutter actor, not a themable object -
-                              // probably something that could be addressed
+const HORIZONTAL_ICON_SIZE = 16; // too bad this can't be defined in theme (cinnamon-app.create_icon_texture returns a clutter actor, not a themable object -
+                                 // probably something that could be addressed
 const ICON_HEIGHT_FACTOR = .64;
+const VERTICAL_ICON_HEIGHT_FACTOR = .75;
 const MAX_TEXT_LENGTH = 1000;
 const FLASH_INTERVAL = 500;
 
@@ -84,15 +86,20 @@ WindowPreview.prototype = {
         Tooltips.TooltipBase.prototype._init.call(this, item.actor);
         this._applet = item._applet;
 
+// FIXME: there is quite a bit of hardcoding in this part of the code, would benefit from setting up separate CSS
+// and getting rid of the hard-coding.
+
         this.actor = new St.Bin({style_class: "switcher-list", style: "margin: 0px; padding: 8px;"});
         this.actor.show_on_set_parent = false;
 
         this.scaleFactor = St.ThemeContext.get_for_stage(global.stage).scale_factor;
 
-        this.actor.set_size(WINDOW_PREVIEW_WIDTH * 1.3 * this.scaleFactor, WINDOW_PREVIEW_HEIGHT * 1.3 * this.scaleFactor);
+        this.actor.set_size(WINDOW_PREVIEW_WIDTH * 1.2 * this.scaleFactor, WINDOW_PREVIEW_HEIGHT * 1.2 * this.scaleFactor);
         Main.uiGroup.add_actor(this.actor);
 
         this.metaWindow = metaWindow;
+        this.muffinWindow = null;
+        this._sizeChangedId = null;
 
         let box = new St.BoxLayout({ vertical: true });
         let hbox = new St.BoxLayout();
@@ -107,7 +114,6 @@ WindowPreview.prototype = {
         this.label = new St.Label();
         this.label.style = "padding: 2px;";
         hbox.add_actor(this.label);
-
         box.add_actor(hbox);
 
         this.thumbnailBin = new St.Bin();
@@ -120,7 +126,7 @@ WindowPreview.prototype = {
         if (this._applet._tooltipShowing)
             this.show();
         else if (!this._showTimer)
-            this._showTimer = Mainloop.timeout_add(300, Lang.bind(this, this._onTimerComplete));
+            this._showTimer = Mainloop.timeout_add(300, Lang.bind(this, this._onShowTimerComplete));
 
         this.mousePosition = event.get_coords();
     },
@@ -134,10 +140,12 @@ WindowPreview.prototype = {
         if (!this.actor || this._applet._menuOpen)
             return
 
-        let muffinWindow = this.metaWindow.get_compositor_private();
-        let windowTexture = muffinWindow.get_texture();
+        this.muffinWindow = this.metaWindow.get_compositor_private();
+        let windowTexture = this.muffinWindow.get_texture();
         let [width, height] = windowTexture.get_size();
-        let scale = Math.min(1.0, WINDOW_PREVIEW_WIDTH / width, WINDOW_PREVIEW_HEIGHT / height);
+        // the 18 is 16 for the icon size + 2px min padding
+        // this is not foolproof - the font used might be large enough to make the label bigger than the icon
+        let scale = Math.min(1.0, WINDOW_PREVIEW_WIDTH / width, (WINDOW_PREVIEW_HEIGHT-18) / height);
 
         if (this.thumbnail) {
             this.thumbnailBin.set_child(null);
@@ -150,6 +158,14 @@ WindowPreview.prototype = {
             height: height * scale * this.scaleFactor
         });
 
+        this._setSize = function() {
+            [width, height] = windowTexture.get_size();
+            scale = Math.min(1.0, WINDOW_PREVIEW_WIDTH / width, (WINDOW_PREVIEW_HEIGHT-18) / height);
+            this.thumbnail.set_size(width * scale * this.scaleFactor, height * scale * this.scaleFactor);
+        };
+        this._sizeChangedId = this.muffinWindow.connect('size-changed',
+            Lang.bind(this, this._setSize));
+
         this.thumbnailBin.set_child(this.thumbnail);
 
         let allocation = this.actor.get_allocation_box();
@@ -158,17 +174,31 @@ WindowPreview.prototype = {
 
         let monitor = Main.layoutManager.findMonitorForActor(this.item);
         let previewTop;
+
         if (this._applet.orientation == St.Side.BOTTOM) {
             previewTop = this.item.get_transformed_position()[1] - previewHeight - 5;
-        } else {
+        } else if (this._applet.orientation == St.Side.TOP) {
             previewTop = this.item.get_transformed_position()[1] + this.item.get_transformed_size()[1] + 5;
+        } else {
+            previewTop = this.item.get_transformed_position()[1];
         }
 
-        let previewLeft = this.item.get_transformed_position()[0] + this.item.get_transformed_size()[0]/2 - previewWidth/2;
+        let previewLeft;
+        if (this._applet.orientation == St.Side.BOTTOM || this._applet.orientation == St.Side.TOP) { 
+            // centre the applet on the window list item if window list is on the top or bottom panel
+            previewLeft = this.item.get_transformed_position()[0] + this.item.get_transformed_size()[0]/2 - previewWidth/2;
+        } else if (this._applet.orientation == St.Side.LEFT) {
+            previewLeft = this.item.get_transformed_position()[0] + this.item.get_transformed_size()[0] + 5;
+        } else if (this._applet.orientation == St.Side.RIGHT) {
+            previewLeft = this.item.get_transformed_position()[0] - previewWidth -5;
+        }
         previewLeft = Math.round(previewLeft);
         previewLeft = Math.max(previewLeft, monitor.x);
         previewLeft = Math.min(previewLeft, monitor.x + monitor.width - previewWidth);
 
+        previewTop  = Math.round(previewTop);
+        previewTop  = Math.min(previewTop, monitor.y + monitor.height - previewHeight);
+
         this.actor.set_position(previewLeft, previewTop);
 
         this.actor.show();
@@ -178,6 +208,10 @@ WindowPreview.prototype = {
     },
 
     hide: function() {
+        if (this._sizeChangedId != null) {
+            this.muffinWindow.disconnect(this._sizeChangedId);
+            this._sizeChangedId = null;
+        }
         if (this.thumbnail) {
             this.thumbnailBin.set_child(null);
             this.thumbnail.destroy();
@@ -193,6 +227,10 @@ WindowPreview.prototype = {
     },
 
     _destroy: function() {
+        if (this._sizeChangedId != null) {
+            this.muffinWindow.disconnect(this._sizeChangedId);
+            this.sizeChangedId = null;
+        }
         if (this.thumbnail) {
             this.thumbnailBin.set_child(null);
             this.thumbnail.destroy();
@@ -210,6 +248,7 @@ function AppMenuButton(applet, metaWindow, alert) {
 
 AppMenuButton.prototype = {
     _init: function(applet, metaWindow, alert) {
+
         this.actor = new Cinnamon.GenericContainer({
             name: 'appMenu',
             style_class: 'window-list-item-box',
@@ -220,11 +259,16 @@ AppMenuButton.prototype = {
         this._applet = applet;
         this.metaWindow = metaWindow;
         this.alert = alert;
+        this.labelVisible = false;
 
         if (this._applet.orientation == St.Side.TOP)
-            this.actor.add_style_class_name('window-list-item-box-top');
-        else
-            this.actor.add_style_class_name('window-list-item-box-bottom');
+            this.actor.add_style_class_name('top');
+        else if (this._applet.orientation == St.Side.BOTTOM)
+            this.actor.add_style_class_name('bottom');
+        else if (this._applet.orientation == St.Side.LEFT)
+            this.actor.add_style_class_name('left');
+        else if (this._applet.orientation == St.Side.RIGHT)
+            this.actor.add_style_class_name('right');
 
         this.actor._delegate = this;
         this.actor.connect('button-release-event', Lang.bind(this, this._onButtonRelease));
@@ -242,9 +286,12 @@ AppMenuButton.prototype = {
         this._iconBox.connect('notify::allocation',
                               Lang.bind(this, this._updateIconBoxClipAndGeometry));
         this.actor.add_actor(this._iconBox);
+
         this._label = new St.Label();
         this.actor.add_actor(this._label);
 
+        this.updateLabelVisible();
+
         this._iconBottomClip = 0;
         this._visible = true;
 
@@ -289,7 +336,7 @@ AppMenuButton.prototype = {
             this._tooltip = new WindowPreview(this, this.metaWindow, this._applet.orientation);
         else
             this._tooltip = new Tooltips.PanelItemTooltip(this, "", this._applet.orientation);
-        
+
         this.setDisplayTitle();
     },
 
@@ -320,7 +367,6 @@ AppMenuButton.prototype = {
                 return item._delegate;
             });
 
-
             windows = windows.reverse();
 
         let i = windows.length;
@@ -344,7 +390,14 @@ AppMenuButton.prototype = {
     },
 
     _onDragBegin: function() {
-        this._draggable._overrideY = this.actor.get_transformed_position()[1];
+        if (this._applet.orientation == St.Side.TOP || this._applet.orientation == St.Side.BOTTOM) {
+            this._draggable._overrideY = this.actor.get_transformed_position()[1];
+            this._draggable._overrideX = null;
+        } else {
+            this._draggable._overrideX = this.actor.get_transformed_position()[0];
+            this._draggable._overrideY = null;
+        }
+
         this._tooltip.hide();
         this._tooltip.preventShow = true;
     },
@@ -362,8 +415,8 @@ AppMenuButton.prototype = {
     },
 
     getDragActor: function() {
-        let clone = new Clutter.Clone({ source: this.actor });
-        clone.width = this.actor.width;
+        let clone    = new Clutter.Clone({ source: this.actor });
+        clone.width  = this.actor.width;
         clone.height = this.actor.height;
         return clone;
     },
@@ -392,15 +445,15 @@ AppMenuButton.prototype = {
     },
 
     setDisplayTitle: function() {
-        let title = this.metaWindow.get_title();
+        let title   = this.metaWindow.get_title();
         let tracker = Cinnamon.WindowTracker.get_default();
         let app = tracker.get_window_app(this.metaWindow);
 
         if (!title) title = app ? app.get_name() : '?';
 
-        /* Sanitize the window title to prevent dodgey window titles such as
+        /* Sanitize the window title to prevent dodgy window titles such as
          * "); DROP TABLE windows; --. Turn all whitespaces into " " because
-         * newline characters are known to cause trouble. Also truncunate the
+         * newline characters are known to cause trouble. Also truncate the
          * title when necessary or else cogl might get unhappy and crash
          * Cinnamon. */
         title = title.replace(/\s/g, " ");
@@ -491,6 +544,10 @@ AppMenuButton.prototype = {
         if (!this.alert && event.get_button() == 3) {
             this.rightClickMenu.mouseEvent = event;
             this.rightClickMenu.toggle();
+
+            if (this._hasFocus()) {
+                this.actor.add_style_pseudo_class('focus');
+            }
         }
     },
 
@@ -499,7 +556,7 @@ AppMenuButton.prototype = {
             Main.activateWindow(this.metaWindow, global.get_current_time());
             this.actor.add_style_pseudo_class('focus');
         } else if (!fromDrag) {
-            this.metaWindow.minimize(global.get_current_time());
+            this.metaWindow.minimize();
             this.actor.remove_style_pseudo_class('focus');
         }
     },
@@ -529,79 +586,124 @@ AppMenuButton.prototype = {
     _getPreferredWidth: function(actor, forHeight, alloc) {
         let [minSize, naturalSize] = this._iconBox.get_preferred_width(forHeight);
         // minimum size just enough for icon if we ever get that many apps going
+
         alloc.min_size = naturalSize + 2 * 3 * global.ui_scale;
 
-        if (this._applet.buttonsUseEntireSpace) {
-            let [lminSize, lnaturalSize] = this._label.get_preferred_width(forHeight);
-            alloc.natural_size = Math.max(150 * global.ui_scale,
-                    lnaturalSize + naturalSize + 3 * 3 * global.ui_scale);
+        if (this._applet.orientation == St.Side.TOP || this._applet.orientation == St.Side.BOTTOM ) {
+        // the 'buttons use entire space' option only makes sense on horizontal panels
+            if (this._applet.buttonsUseEntireSpace) {
+                let [lminSize, lnaturalSize] = this._label.get_preferred_width(forHeight);
+                alloc.natural_size = Math.max(150 * global.ui_scale,
+                        lnaturalSize + naturalSize + 3 * 3 * global.ui_scale);
+            } else {
+                alloc.natural_size = 150 * global.ui_scale;
+            }
         } else {
-            alloc.natural_size = 150 * global.ui_scale;
+            alloc.natural_size = this._applet._panelHeight;
         }
     },
 
     _getPreferredHeight: function(actor, forWidth, alloc) {
         let [minSize1, naturalSize1] = this._iconBox.get_preferred_height(forWidth);
-        let [minSize2, naturalSize2] = this._label.get_preferred_height(forWidth);
 
-        alloc.min_size = Math.max(minSize1, minSize2);
-        alloc.natural_size = Math.max(naturalSize1, naturalSize2);
+        if (this.labelVisible) {
+            let [minSize2, naturalSize2] = this._label.get_preferred_height(forWidth);
+            alloc.min_size = Math.max(minSize1, minSize2);
+        } else {
+            alloc.min_size = minSize1;
+        }
+
+        if (this._applet.orientation == St.Side.TOP || this._applet.orientation == St.Side.BOTTOM ) {
+            /* putting a container around the actor for layout management reasons affects the allocation,
+               causing the visible border to pull in close around the contents which is not the desired
+               (pre-existing) behaviour, so need to push the visible border back towards the panel edge.
+               Assigning the natural size to the full panel height used to cause recursion errors but seems fine now.
+               If this happens to avoid this you can subtract 1 or 2 pixels, but this will give an unreactive
+               strip at the edge of the screen */
+            alloc.natural_size = this._applet._panelHeight;
+        } else {
+            alloc.natural_size = naturalSize1;
+        }
     },
 
     _allocate: function(actor, box, flags) {
         let allocWidth = box.x2 - box.x1;
         let allocHeight = box.y2 - box.y1;
+
         let childBox = new Clutter.ActorBox();
 
         let [minWidth, minHeight, naturalWidth, naturalHeight] = this._iconBox.get_preferred_size();
 
         let direction = this.actor.get_text_direction();
-
         let xPadding = 3 * global.ui_scale;
         let yPadding = Math.floor(Math.max(0, allocHeight - naturalHeight) / 2);
+
         childBox.y1 = yPadding;
         childBox.y2 = childBox.y1 + Math.min(naturalHeight, allocHeight);
-        if (direction == Clutter.TextDirection.LTR) {
-            if (allocWidth < naturalWidth + xPadding * 2)
-                childBox.x1 = Math.max(0, (allocWidth - naturalWidth) / 2)
-            else
-                childBox.x1 = Math.min(allocWidth, xPadding);
-            childBox.x2 = Math.min(childBox.x1 + naturalWidth, allocWidth);
+
+        if (this.labelVisible) {
+            if (direction == Clutter.TextDirection.LTR) {
+                if (allocWidth < naturalWidth + xPadding * 2)
+                    childBox.x1 = Math.max(0, (allocWidth - naturalWidth) / 2);
+                else
+                    childBox.x1 = Math.min(allocWidth, xPadding);
+                childBox.x2 = Math.min(childBox.x1 + naturalWidth, allocWidth);
+            } else {
+                if (allocWidth < naturalWidth + xPadding * 2)
+                    childBox.x1 = Math.max(0, (allocWidth - naturalWidth) / 2);
+                else
+                    childBox.x1 = allocWidth - naturalWidth - xPadding;
+                childBox.x2 = Math.min(childBox.x1 + naturalWidth, allocWidth);
+            }
         } else {
-            if (allocWidth < naturalWidth + xPadding * 2)
-                childBox.x1 = Math.max(0, (allocWidth - naturalWidth)/2);
+            if (allocWidth < naturalWidth)
+                childBox.x1 = Math.max(0, (allocWidth - naturalWidth) / 2);
             else
-                childBox.x1 = allocWidth - naturalWidth - xPadding;
+                childBox.x1 = (allocWidth - naturalWidth) / 2;
             childBox.x2 = Math.min(childBox.x1 + naturalWidth, allocWidth);
         }
+
         this._iconBox.allocate(childBox, flags);
 
-        let remX1 = childBox.x2, remX2 = allocWidth;
+        if (this.labelVisible) {
+            [minWidth, minHeight, naturalWidth, naturalHeight] = this._label.get_preferred_size();
 
-        [minWidth, minHeight, naturalWidth, naturalHeight] = this._label.get_preferred_size();
+            yPadding = Math.floor(Math.max(0, allocHeight - naturalHeight) / 2);
+            childBox.y1 = yPadding;
+            childBox.y2 = childBox.y1 + Math.min(naturalHeight, allocHeight);
+            if (direction == Clutter.TextDirection.LTR) {
+                // Reuse the values from the previous allocation
+                childBox.x1 = Math.min(childBox.x2 + xPadding, Math.max(0, allocWidth - xPadding));
+                childBox.x2 = Math.max(childBox.x1, allocWidth - xPadding);
+            } else {
+                childBox.x2 = Math.max(childBox.x1 - xPadding, 0);
+                childBox.x1 = Math.min(childBox.x2, xPadding);
+            }
 
-        yPadding = Math.floor(Math.max(0, allocHeight - naturalHeight) / 2);
-        childBox.y1 = yPadding;
-        childBox.y2 = childBox.y1 + Math.min(naturalHeight, allocHeight);
-        if (direction == Clutter.TextDirection.LTR) {
-            // Reuse the values from the previous allocation
-            childBox.x1 = Math.min(childBox.x2 + xPadding, Math.max(0, allocWidth - xPadding));
-            childBox.x2 = Math.max(childBox.x1, allocWidth - xPadding);
+            this._label.allocate(childBox, flags);
+        }
+    },
+
+    updateLabelVisible: function() {
+        if (this._applet.orientation == St.Side.TOP || this._applet.orientation == St.Side.BOTTOM) {
+            this._label.show();
+            this.labelVisible = true;
         } else {
-            childBox.x2 = Math.max(childBox.x1 - xPadding, 0);
-            childBox.x1 = Math.min(childBox.x2, xPadding);
+            this._label.hide();
+            this.labelVisible = false;
         }
-        this._label.allocate(childBox, flags);
     },
 
     setIcon: function() {
         let tracker = Cinnamon.WindowTracker.get_default();
         let app = tracker.get_window_app(this.metaWindow);
 
-        if (this._applet._scaleMode)
+        if (this._applet._scaleMode && this.labelVisible)
             this.iconSize = Math.round(this._applet._panelHeight * ICON_HEIGHT_FACTOR / global.ui_scale);
+        else if (!this.labelVisible)
+            this.iconSize = Math.round(this._applet._panelHeight * VERTICAL_ICON_HEIGHT_FACTOR / global.ui_scale);
         else
-            this.iconSize = DEFAULT_ICON_SIZE;
+            this.iconSize = HORIZONTAL_ICON_SIZE;
 
         let icon = app ?
             app.create_icon_texture(this.iconSize) :
@@ -662,19 +764,10 @@ AppMenuButtonRightClickMenu.prototype = {
         this.metaWindow = metaWindow;
     },
 
-    _populateMenu: function(){
-        this.box.pack_start = this.orientation == St.Side.TOP;
-
+    _populateMenu: function() {
         let mw = this.metaWindow;
         let item;
         let length;
-        item = new PopupMenu.PopupIconMenuItem(_("Configure the window list"), "system-run", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this, function() {
-            Util.spawnCommandLine("cinnamon-settings applets window-list@cinnamon.org " + this._launcher._applet.instance_id);
-        }));
-        this.addMenuItem(item);
-
-        this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
 
         // Move to monitor
         if ((length = Main.layoutManager.monitors.length) == 2) {
@@ -718,39 +811,60 @@ AppMenuButtonRightClickMenu.prototype = {
 
                 let curr_index = mw.get_workspace().index();
                 for (let i = 0; i < length; i++) {
-                    if (i == curr_index) continue;
-
                     // Make the index a local variable to pass to function
                     let j = i;
-                    item.menu.addAction(
-                            Main.workspace_names[i] ? Main.workspace_names[i] : Main._makeDefaultWorkspaceName(i),
-                            function() {
-                                mw.change_workspace(global.screen.get_workspace_by_index(j));
-                            });
+                    let name = Main.workspace_names[i] ? Main.workspace_names[i] : Main._makeDefaultWorkspaceName(i);
+                    let ws = new PopupMenu.PopupMenuItem(name);
+
+                    if (i == curr_index)
+                        ws.setSensitive(false);
+
+                    ws.connect('activate', function() {
+                       mw.change_workspace(global.screen.get_workspace_by_index(j));
+                    });
+                    item.menu.addMenuItem(ws);
                 }
 
             }
         }
 
+        // Preferences
+        let subMenu = new PopupMenu.PopupSubMenuMenuItem(_("Preferences"));
+        this.addMenuItem(subMenu);
+
+        item = new PopupMenu.PopupIconMenuItem(_("About..."), "dialog-question", St.IconType.SYMBOLIC);
+        item.connect('activate', Lang.bind(this._launcher._applet, this._launcher._applet.openAbout));
+        subMenu.menu.addMenuItem(item);
+
+        item = new PopupMenu.PopupIconMenuItem(_("Configure..."), "system-run", St.IconType.SYMBOLIC);
+        item.connect('activate', Lang.bind(this._launcher._applet, this._launcher._applet.configureApplet));
+        subMenu.menu.addMenuItem(item);
+
+        item = new PopupMenu.PopupIconMenuItem(_("Remove 'Window list'"), "edit-delete", St.IconType.SYMBOLIC);
+        item.connect('activate', Lang.bind(this, function() {
+            AppletManager._removeAppletFromPanel(this._launcher._applet._uuid, this._launcher._applet.instance_id);
+        }));
+        subMenu.menu.addMenuItem(item);
+
         this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
 
         // Close all/others
-        item = new PopupMenu.PopupMenuItem(_("Close all"));
+        item = new PopupMenu.PopupIconMenuItem(_("Close all"), "application-exit", St.IconType.SYMBOLIC);
         item.connect('activate', Lang.bind(this, function() {
             for (let window of this._windows)
                 if (window.actor.visible &&
                    !window._needsAttention)
-                    window.metaWindow.delete(global.get_current_time);
+                    window.metaWindow.delete(global.get_current_time());
         }));
         this.addMenuItem(item);
 
-        item = new PopupMenu.PopupMenuItem(_("Close others"));
+        item = new PopupMenu.PopupIconMenuItem(_("Close others"), "window-close", St.IconType.SYMBOLIC);
         item.connect('activate', Lang.bind(this, function() {
             for (let window of this._windows)
                 if (window.actor.visible &&
                     window.metaWindow != this.metaWindow &&
                    !window._needsAttention)
-                    window.metaWindow.delete(global.get_current_time);
+                    window.metaWindow.delete(global.get_current_time());
         }));
         this.addMenuItem(item);
 
@@ -766,12 +880,12 @@ AppMenuButtonRightClickMenu.prototype = {
         }
 
         if (mw.minimized) {
-            item = new PopupMenu.PopupMenuItem(_("Restore"));
+            item = new PopupMenu.PopupIconMenuItem(_("Restore"), "view-sort-descending", St.IconType.SYMBOLIC);
             item.connect('activate', function() {
                 Main.activateWindow(mw, global.get_current_time());
             });
         } else {
-            item = new PopupMenu.PopupMenuItem(_("Minimize"));
+            item = new PopupMenu.PopupIconMenuItem(_("Minimize"), "view-sort-ascending", St.IconType.SYMBOLIC);
             item.connect('activate', function() {
                 mw.minimize(global.get_current_time());
             });
@@ -779,19 +893,19 @@ AppMenuButtonRightClickMenu.prototype = {
         this.addMenuItem(item);
 
         if (mw.get_maximized()) {
-            item = new PopupMenu.PopupMenuItem(_("Unmaximize"));
+            item = new PopupMenu.PopupIconMenuItem(_("Unmaximize"), "view-restore", St.IconType.SYMBOLIC);
             item.connect('activate', function() {
                 mw.unmaximize(Meta.MaximizeFlags.HORIZONTAL | Meta.MaximizeFlags.VERTICAL);
             });
         } else {
-            item = new PopupMenu.PopupMenuItem(_("Maximize"));
+            item = new PopupMenu.PopupIconMenuItem(_("Maximize"), "view-fullscreen", St.IconType.SYMBOLIC);
             item.connect('activate', function() {
                 mw.maximize(Meta.MaximizeFlags.HORIZONTAL | Meta.MaximizeFlags.VERTICAL);
             });
         }
         this.addMenuItem(item);
 
-        item = new PopupMenu.PopupMenuItem(_("Close"));
+        item = new PopupMenu.PopupIconMenuItem(_("Close"), "edit-delete", St.IconType.SYMBOLIC);
         item.connect('activate', function() {
             mw.delete(global.get_current_time());
         });
@@ -822,10 +936,27 @@ MyApplet.prototype = {
     _init: function(orientation, panel_height, instance_id) {
         Applet.Applet.prototype._init.call(this, orientation, panel_height, instance_id);
 
-        this.actor.set_track_hover(false);
-        this.actor.add_style_class_name("window-list-box");
+        this.setAllowedLayout(Applet.AllowedLayout.BOTH);
 
+        this.actor.set_track_hover(false);
+        this.actor.set_style_class_name("window-list-box");
         this.orientation = orientation;
+        this.appletEnabled = false;
+        //
+        // A layout manager is used to cater for vertical panels as well as horizontal
+        //
+        let manager;
+        if (this.orientation == St.Side.TOP || this.orientation == St.Side.BOTTOM) {
+            manager = new Clutter.BoxLayout( { orientation: Clutter.Orientation.HORIZONTAL });
+        } else {
+            manager = new Clutter.BoxLayout( { orientation: Clutter.Orientation.VERTICAL });
+            this.actor.add_style_class_name("vertical");
+        }
+
+        this.manager = manager;
+        this.manager_container = new Clutter.Actor( { layout_manager: manager } );
+        this.actor.add_actor (this.manager_container);
+
         this.dragInProgress = false;
         this._tooltipShowing = false;
         this._tooltipErodeTimer = null;
@@ -836,32 +967,12 @@ MyApplet.prototype = {
 
         this.settings = new Settings.AppletSettings(this, "window-list@cinnamon.org", this.instance_id);
 
-        this.settings.bindProperty(Settings.BindingDirection.IN,
-                "enable-alerts",
-                "enableAlerts",
-                this._updateAttentionGrabber,
-                null);
-        this.settings.bindProperty(Settings.BindingDirection.IN,
-                "enable-scrolling",
-                "scrollable",
-                this._onEnableScrollChanged,
-                null);
-        this.settings.bindProperty(Settings.BindingDirection.IN,
-                "reverse-scrolling",
-                "reverseScroll",
-                null, null);
-        this.settings.bindProperty(Settings.BindingDirection.IN,
-                "middle-click-close",
-                "middleClickClose",
-                null, null);
-        this.settings.bindProperty(Settings.BindingDirection.IN,
-                "buttons-use-entire-space",
-                "buttonsUseEntireSpace",
-                this._refreshItems, null);
-        this.settings.bindProperty(Settings.BindingDirection.IN,
-                "window-preview",
-                "usePreview",
-                this._onPreviewChanged, null);
+        this.settings.bind("enable-alerts", "enableAlerts", this._updateAttentionGrabber);
+        this.settings.bind("enable-scrolling", "scrollable", this._onEnableScrollChanged);
+        this.settings.bind("reverse-scrolling", "reverseScroll");
+        this.settings.bind("middle-click-close", "middleClickClose");
+        this.settings.bind("buttons-use-entire-space", "buttonsUseEntireSpace", this._refreshAllItems);
+        this.settings.bind("window-preview", "usePreview", this._onPreviewChanged);
 
         this.signals = new SignalManager.SignalManager(this);
 
@@ -871,6 +982,7 @@ MyApplet.prototype = {
         this.signals.connect(global.screen, 'window-removed', this._onWindowRemoved);
         this.signals.connect(global.screen, 'window-monitor-changed', this._onWindowMonitorChanged);
         this.signals.connect(global.screen, 'window-workspace-changed', this._onWindowWorkspaceChanged);
+        this.signals.connect(global.screen, 'window-skip-taskbar-changed', this._onWindowSkipTaskbarChanged);
         this.signals.connect(global.screen, 'monitors-changed', this._updateWatchedMonitors);
         this.signals.connect(global.window_manager, 'switch-workspace', this._refreshAllItems);
 
@@ -880,12 +992,19 @@ MyApplet.prototype = {
         this.signals.connect(global.window_manager, 'map', this._onWindowStateChange);
         this.signals.connect(global.window_manager, 'tile', this._onWindowStateChange);
 
+        this.actor.connect('style-changed', Lang.bind(this, this._updateSpacing));
+
         global.settings.bind("panel-edit-mode", this.actor, "reactive", Gio.SettingsBindFlags.DEFAULT);
 
         this.on_orientation_changed(orientation);
         this._updateAttentionGrabber();
     },
 
+    on_applet_added_to_panel: function(userEnabled) {
+        this._updateSpacing();
+        this.appletEnabled = true;
+    },
+
     on_applet_removed_from_panel: function() {
         this.signals.disconnectAllSignals();
     },
@@ -900,24 +1019,66 @@ MyApplet.prototype = {
 
     on_orientation_changed: function(orientation) {
         this.orientation = orientation;
+
+        for (let window of this._windows)
+            window.updateLabelVisible();
+
+        if (orientation == St.Side.TOP || orientation == St.Side.BOTTOM) {
+            this.manager.set_vertical(false);
+            this._reTitleItems();
+            this.actor.remove_style_class_name("vertical");
+        } else {
+            this.manager.set_vertical(true);
+            this.actor.add_style_class_name("vertical");
+            this.actor.set_x_align(Clutter.ActorAlign.CENTER);
+            this.actor.set_important(true);
+        }
+
+        // Any padding/margin is removed on one side so that the AppMenuButton
+        // boxes butt up against the edge of the screen
+
         if (orientation == St.Side.TOP) {
-            for (let child of this.actor.get_children()) {
-                child.set_style_class_name('window-list-item-box window-list-box-top');
+            for (let child of this.manager_container.get_children()) {
+                child.set_style_class_name('window-list-item-box top');
                 child.set_style('margin-top: 0px; padding-top: 0px;');
             }
             this.actor.set_style('margin-top: 0px; padding-top: 0px;');
-        } else {
-            for (let child of this.actor.get_children()) {
-                child.set_style_class_name('window-list-item-box window-list-box-bottom');
+        } else if (orientation == St.Side.BOTTOM) {
+            for (let child of this.manager_container.get_children()) {
+                child.set_style_class_name('window-list-item-box bottom');
                 child.set_style('margin-bottom: 0px; padding-bottom: 0px;');
             }
             this.actor.set_style('margin-bottom: 0px; padding-bottom: 0px;');
+        } else if (orientation == St.Side.LEFT) {
+            for (let child of this.manager_container.get_children()) {
+                child.set_style_class_name('window-list-item-box left');
+                child.set_style('margin-left 0px; padding-left: 0px; padding-right: 0px; margin-right: 0px;');
+                child.set_x_align(Clutter.ActorAlign.CENTER);
+            }
+            this.actor.set_style('margin-left: 0px; padding-left: 0px; padding-right: 0px; margin-right: 0px;');
+        } else if (orientation == St.Side.RIGHT) {
+            for (let child of this.manager_container.get_children()) {
+                child.set_style_class_name('window-list-item-box right');
+                child.set_style('margin-left: 0px; padding-left: 0px; padding-right: 0px; margin-right: 0px;');
+                child.set_x_align(Clutter.ActorAlign.CENTER);
+            }
+            this.actor.set_style('margin-right: 0px; padding-right: 0px; padding-left: 0px; margin-left: 0px;');
         }
+
+        if (this.appletEnabled) {
+            this._updateSpacing();
+        }
+    },
+
+    _updateSpacing: function() {
+        let themeNode = this.actor.get_theme_node();
+        let spacing = themeNode.get_length('spacing');
+        this.manager.set_spacing(spacing * global.ui_scale);
     },
 
     _onWindowAdded: function(screen, metaWindow, monitor) {
         if (this._shouldAdd(metaWindow))
-            this._addWindow(metaWindow);
+            this._addWindow(metaWindow, false);
     },
 
     _onWindowRemoved: function(screen, metaWindow) {
@@ -926,7 +1087,7 @@ MyApplet.prototype = {
 
     _onWindowMonitorChanged: function(screen, metaWindow, monitor) {
         if (this._shouldAdd(metaWindow))
-            this._addWindow(metaWindow);
+            this._addWindow(metaWindow, false);
         else
             this._removeWindow(metaWindow);
     },
@@ -938,6 +1099,17 @@ MyApplet.prototype = {
             this._refreshItem(window);
     },
 
+    _onWindowSkipTaskbarChanged: function(screen, metaWindow) {
+        let window = this._windows.find(win => (win.metaWindow == metaWindow));
+
+        if (window && !Main.isInteresting(metaWindow)) {
+            this._removeWindow(metaWindow);
+            return;
+        }
+
+        this._onWindowAdded(screen, metaWindow, 0);
+    },
+
     _updateAttentionGrabber: function() {
         if (this.enableAlerts) {
             this.signals.connect(global.display, "window-marked-urgent", this._onWindowDemandsAttention);
@@ -1000,6 +1172,12 @@ MyApplet.prototype = {
         this._onFocus();
     },
 
+    _reTitleItems: function() {
+        for (let window of this._windows) {
+            window.setDisplayTitle();
+        }
+    },
+
     _onWindowStateChange: function(cinnamonwm, actor) {
         for (let window of this._windows)
             if (window.metaWindow == actor.metaWindow)
@@ -1047,7 +1225,7 @@ MyApplet.prototype = {
 
         for (let window of windows) {
             if (this._shouldAdd(window))
-                this._addWindow(window);
+                this._addWindow(window, false);
             else
                 this._removeWindow(window);
         }
@@ -1056,11 +1234,11 @@ MyApplet.prototype = {
     _addWindow: function(metaWindow, alert) {
         for (let window of this._windows)
             if (window.metaWindow == metaWindow &&
-                window.temp == alert)
+                window.alert == alert)
                 return;
 
         let appButton = new AppMenuButton(this, metaWindow, alert);
-        this.actor.add(appButton.actor);
+        this.manager_container.add_actor(appButton.actor);
 
         this._windows.push(appButton);
         /* We want to make the AppMenuButtons look like they are ordered by
@@ -1069,7 +1247,7 @@ MyApplet.prototype = {
          * default, so move it to the start if needed */
         if (alert) {
             if (metaWindow.get_workspace().index() < global.screen.get_active_workspace_index())
-                this.actor.set_child_at_index(appButton.actor, 0);
+                this.manager_container.set_child_at_index(appButton.actor, 0);
         } else {
             if (metaWindow.get_workspace() != global.screen.get_active_workspace())
                 appButton.actor.hide();
@@ -1099,10 +1277,14 @@ MyApplet.prototype = {
             return DND.DragMotionResult.NO_DROP;
 
         source.actor.hide();
-        let children = this.actor.get_children();
+        let children = this.manager_container.get_children();
 
         let pos = children.length;
-        while (--pos && x < children[pos].get_allocation_box().x1);
+
+        if (this.manager_container.height > this.manager_container.width) // assume oriented vertically
+            while (--pos && y < children[pos].get_allocation_box().y1);
+        else
+            while (--pos && x < children[pos].get_allocation_box().x1);
 
         this._dragPlaceholderPos = pos;
 
@@ -1110,11 +1292,12 @@ MyApplet.prototype = {
             this._dragPlaceholder = new DND.GenericDragPlaceholderItem();
             this._dragPlaceholder.child.set_width (source.actor.width);
             this._dragPlaceholder.child.set_height (source.actor.height);
-            this.actor.insert_child_at_index(this._dragPlaceholder.actor,
-                                             this._dragPlaceholderPos);
+
+            this.manager_container.insert_child_at_index(this._dragPlaceholder.actor,
+                                                         this._dragPlaceholderPos);
         } else {
-            this.actor.set_child_at_index(this._dragPlaceholder.actor,
-                    this._dragPlaceholderPos);
+            this.manager_container.set_child_at_index(this._dragPlaceholder.actor,
+                                                         this._dragPlaceholderPos);
         }
 
         return DND.DragMotionResult.MOVE_DROP;
@@ -1124,7 +1307,7 @@ MyApplet.prototype = {
         if (!(source instanceof AppMenuButton)) return false;
         if (this._dragPlaceholderPos == undefined) return false;
 
-        this.actor.set_child_at_index(source.actor, this._dragPlaceholderPos);
+        this.manager_container.set_child_at_index(source.actor, this._dragPlaceholderPos);
 
         return true;
     },
