@@ -119,15 +119,14 @@ const Settings = imports.ui.settings;
 const Systray = imports.ui.systray;
 const Accessibility = imports.ui.accessibility;
 
-const DEFAULT_BACKGROUND_COLOR = new Clutter.Color();
-DEFAULT_BACKGROUND_COLOR.from_pixel(0x2266bbff);
-
 const LAYOUT_TRADITIONAL = "traditional";
 const LAYOUT_FLIPPED = "flipped";
 const LAYOUT_CLASSIC = "classic";
 
 const CIN_LOG_FOLDER = GLib.get_home_dir() + '/.cinnamon/';
 
+let DEFAULT_BACKGROUND_COLOR = Clutter.Color.from_pixel(0x2266bbff);
+
 let panel = null;
 let soundManager = null;
 let backgroundManager = null;
@@ -221,6 +220,7 @@ function _initUserSession() {
     global.screen.override_workspace_layout(Meta.ScreenCorner.TOPLEFT, false, 1, -1);
 
     systrayManager = new Systray.SystrayManager();
+    indicatorManager = new IndicatorManager.IndicatorManager();
     
     ExtensionSystem.init();
 
@@ -271,7 +271,7 @@ function start() {
                     log_dir.make_directory_with_parents(null);
                 lg_log_file = log_filename.append_to(0, null);
             } else {
-                log_filename.copy(log_backup_filename, 1, null, null, null);
+                log_filename.copy(log_backup_filename, 1, null, null);
                 log_filename.delete(null);
                 lg_log_file = log_filename.append_to(0, null);
             }
@@ -367,6 +367,8 @@ function start() {
 
     global.reparentActor(global.top_window_group, global.stage);
 
+    global.menuStackLength = 0;
+
     layoutManager = new Layout.LayoutManager();
 
     Panel.checkPanelUpgrade();
@@ -399,7 +401,6 @@ function start() {
     
     wm = new WindowManager.WindowManager();
     messageTray = new MessageTray.MessageTray();
-    indicatorManager = new IndicatorManager.IndicatorManager();
     keyboard = new Keyboard.Keyboard();
     notificationDaemon = new NotificationDaemon.NotificationDaemon();
     windowAttentionHandler = new WindowAttentionHandler.WindowAttentionHandler();
@@ -894,7 +895,7 @@ function warningNotify(msg, details, icon) {
     messageTray.add(source);
     let notification = new MessageTray.Notification(source, msg, details, { icon: icon });
     notification.setTransient(false);
-    notification.setUrgency(MessageTray.Urgency.WARNING);
+    notification.setUrgency(MessageTray.Urgency.HIGH);
     source.notify(notification);
 }
 
@@ -926,7 +927,22 @@ function notifyError(msg, details) {
  * extension system as well as debugging.
  */
 function _log(category, msg) {
-    let text = msg;
+    if (msg == undefined) {
+        _log('error', _('logging failed: message was null or undefined'));
+        return;
+    }
+
+    let cat, text;
+    if (typeof category !== 'string')
+        cat = 'info';
+    else
+        cat = category;
+
+    if (typeof msg !== 'string')
+        text = msg.toString();
+    else
+        text = msg;
+
     if (arguments.length > 2) {
         text += ': ';
         for (let i = 2; i < arguments.length; i++) {
@@ -936,7 +952,7 @@ function _log(category, msg) {
         }
     }
     let out = {timestamp: new Date().getTime().toString(),
-                         category: category,
+                         category: cat,
                          message: text };
     _errorLogStack.push(out);
     if (lookingGlass)
@@ -953,7 +969,23 @@ function _log(category, msg) {
  * Returns (boolean): whether @obj is an error object
  */
 function isError(obj) {
-    return typeof(obj) == 'object' && 'message' in obj && 'stack' in obj;
+    if (obj == undefined) return false;
+
+    let isErr = false;
+    if (typeof(obj) == 'object' && 'message' in obj && 'stack' in obj) {
+        isErr = true;
+    } else if (obj instanceof GLib.Error) {
+        // Make existing logging functionality work as expected when passed
+        // a GLib.Error which doesn't normally have a stack trace attached.
+        let stack = new Error().stack;
+        // This is reached the first time isError is called by a _log function,
+        // so strip off this function call and the _log function that called us.
+        let strPos = stack.indexOf('\n', stack.indexOf('\n') + 1)  + 1;
+        stack = stack.substr(strPos);
+        obj.stack = stack;
+        isErr = true;
+    }
+    return isErr;
 }
 
 /**
